(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
  typeof define === 'function' && define.amd ? define(['exports'], factory) :
  (factory((global.Unistore = global.Unistore || {})));
}(this, function (exports) { 'use strict';

  var babelHelpers = {};
  babelHelpers.typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) {
    return typeof obj;
  } : function (obj) {
    return obj && typeof Symbol === "function" && obj.constructor === Symbol ? "symbol" : typeof obj;
  };

  babelHelpers.classCallCheck = function (instance, Constructor) {
    if (!(instance instanceof Constructor)) {
      throw new TypeError("Cannot call a class as a function");
    }
  };

  babelHelpers.inherits = function (subClass, superClass) {
    if (typeof superClass !== "function" && superClass !== null) {
      throw new TypeError("Super expression must either be null or a function, not " + typeof superClass);
    }

    subClass.prototype = Object.create(superClass && superClass.prototype, {
      constructor: {
        value: subClass,
        enumerable: false,
        writable: true,
        configurable: true
      }
    });
    if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass;
  };

  babelHelpers.possibleConstructorReturn = function (self, call) {
    if (!self) {
      throw new ReferenceError("this hasn't been initialised - super() hasn't been called");
    }

    return call && (typeof call === "object" || typeof call === "function") ? call : self;
  };

  babelHelpers;

  // https://github.com/sindresorhus/object-assign

  var hasOwnProperty = Object.prototype.hasOwnProperty;

  function toObject(val) {
    if (val === null || val === undefined) {
      throw new TypeError('Object.assign cannot be called with null or undefined');
    }

    return Object(val);
  }

  var objectAssign = Object.assign || function (target, source) {
    var to = toObject(target);
    var from = undefined;

    for (var s = 1, len = arguments.length; s < len; s++) {
      from = Object(arguments[s]);

      for (var key in from) {
        if (hasOwnProperty.call(from, key)) {
          to[key] = from[key];
        }
      }
    }

    return to;
  };

  function clone(object) {
    return objectAssign({}, object);
  }

  var RESTAdapter = function () {
    function RESTAdapter() {
      babelHelpers.classCallCheck(this, RESTAdapter);

      this._endpointMap = {};
      if (this.constructor.name === 'RESTAdapter') throw TypeError('RESTAdapter is a base class');
    }

    RESTAdapter.prototype.registerEndpoints = function registerEndpoints(type, endpoints) {
      this._endpointMap[type] = endpoints;
      return this;
    };

    RESTAdapter.prototype.isSuccess = function isSuccess(status) {
      return !!(status && status < 400);
    };

    RESTAdapter.prototype.endpointFor = function endpointFor(type, params) {
      var typeEndpoints = this._endpointMap[type];

      if (typeEndpoints) {
        var endpointOne = typeEndpoints.one;
        var endpointMany = typeEndpoints.many;
        var typeofParams = typeof params === 'undefined' ? 'undefined' : babelHelpers.typeof(params);
        var dynamicSegments = undefined,
            keySegment = undefined,
            keyName = undefined,
            keyValue = undefined;

        if (typeofParams === 'undefined') {
          return endpointMany;
        }

        if (endpointOne) {
          dynamicSegments = endpointOne.match(/:\w+/g);
          if (dynamicSegments) {
            keySegment = dynamicSegments[dynamicSegments.length - 1];
            keyName = keySegment.slice(1);
            keyValue = params[keyName];
          }
        }

        if (typeofParams === 'object') {
          if (keyValue === undefined) {
            return this.buildUrl(endpointMany, params);
          }

          delete params[keyName];
          return this.buildUrl(endpointOne.replace(keySegment, keyValue), params);
        }

        return endpointOne.replace(keySegment, params);
      }
    };

    RESTAdapter.prototype.buildUrl = function buildUrl(url, data) {
      var query = [];
      for (var prop in data) {
        if (data.hasOwnProperty(prop)) {
          query.push(encodeURIComponent(prop) + '=' + encodeURIComponent(data[prop]));
        }
      }

      if (query.length) {
        var indexOfQuestionMark = url.indexOf('?');
        if (indexOfQuestionMark === -1) {
          url += '?';
        } else if (indexOfQuestionMark !== url.length - 1) {
          url += '&';
        }

        url += query.join('&');
      }
      return url;
    };

    return RESTAdapter;
  }();

  var FetchAdapter = function (_RESTAdapter) {
    babelHelpers.inherits(FetchAdapter, _RESTAdapter);

    function FetchAdapter() {
      babelHelpers.classCallCheck(this, FetchAdapter);
      return babelHelpers.possibleConstructorReturn(this, _RESTAdapter.apply(this, arguments));
    }

    FetchAdapter.prototype.get = function get(type, params) {
      var _this2 = this;

      var url = this.endpointFor(type, params);

      return fetch(url).then(function (response) {
        return _this2.isSuccess(response.status) ? response.json() : Promise.reject(response);
      });
    };

    return FetchAdapter;
  }(RESTAdapter);

  var XHRAdapter = function (_RESTAdapter) {
    babelHelpers.inherits(XHRAdapter, _RESTAdapter);

    function XHRAdapter() {
      babelHelpers.classCallCheck(this, XHRAdapter);
      return babelHelpers.possibleConstructorReturn(this, _RESTAdapter.apply(this, arguments));
    }

    XHRAdapter.prototype.get = function get(type, params) {
      var _this2 = this;

      return new Promise(function (resolve, reject) {
        var url = _this2.endpointFor(type, params);
        var xhr = new XMLHttpRequest();

        xhr.onreadystatechange = function () {
          if (xhr.readyState === 4) {
            if (_this2.isSuccess(xhr.status)) {
              try {
                resolve(JSON.parse(xhr.responseText));
              } catch (e) {
                reject(e);
              }
            } else {
              reject(xhr);
            }
          }
        };

        xhr.open('GET', url);
        xhr.send();
      });
    };

    return XHRAdapter;
  }(RESTAdapter);

  var http = undefined;
  var https = undefined;
  if (typeof window === 'undefined' && typeof require !== 'undefined') {
    http = require('http');
    https = require('https');
  }

  function httpProtocolModuleForUrl(url) {
    return url.match(/^(?:http(s?)\:\/\/)?/i)[0] === 'https://' ? https : http;
  }

  var HTTPAdapter = function (_RESTAdapter) {
    babelHelpers.inherits(HTTPAdapter, _RESTAdapter);

    function HTTPAdapter() {
      babelHelpers.classCallCheck(this, HTTPAdapter);
      return babelHelpers.possibleConstructorReturn(this, _RESTAdapter.apply(this, arguments));
    }

    HTTPAdapter.prototype.get = function get(type, params) {
      var _this2 = this;

      return new Promise(function (resolve, reject) {
        var url = _this2.endpointFor(type, params);

        httpProtocolModuleForUrl(url).get(url, function (response) {
          var status = response.statusCode;
          var body = '';

          // normalizes 'status' property with other adapters
          response.status = status;

          if (_this2.isSuccess(status)) {
            response.on('data', function (data) {
              body += data;
            });
            response.on('end', function () {
              try {
                resolve(JSON.parse(body));
              } catch (e) {
                reject(e);
              }
            });
          } else {
            reject(response);
          }
        }).on('error', reject);
      });
    };

    return HTTPAdapter;
  }(RESTAdapter);

  function autoAdapter() {
    var win = typeof window !== 'undefined' && window;

    if (win) {
      if (typeof win.fetch !== 'undefined') {
        return new FetchAdapter();
      }
      return new XHRAdapter();
    }

    return new HTTPAdapter();
  }

  var Store = function () {
    function Store() {
      var opts = arguments.length <= 0 || arguments[0] === undefined ? {} : arguments[0];
      babelHelpers.classCallCheck(this, Store);

      this.primaryKey = opts.primaryKey || 'id';
      this.adapter = opts.adapter || autoAdapter();
      this._registry = {};
      this._data = {};
      this._transforms = [[], []];
    }

    Store.prototype.register = function register(type) {
      var opts = arguments.length <= 1 || arguments[1] === undefined ? {} : arguments[1];

      var data = this._data;
      var endpoints = opts.endpoints;

      this._registry[type] = opts;
      data[type] = data[type] || {};
      endpoints && this.adapter.registerEndpoints(type, endpoints);
      return this;
    };

    Store.prototype.lookup = function lookup(type) {
      return this._registry[type];
    };

    Store.prototype.clear = function clear() {
      var data = this._data;
      for (var type in data) {
        if (data.hasOwnProperty(type)) {
          data[type] = {};
        }
      }
    };

    Store.prototype.get = function get(type, params) {
      var fnc = this._isManyRequest(type, params) ? '_getMany' : '_getByKey';
      return this[fnc](type, params);
    };

    Store.prototype.push = function push(type, records) {
      var fnc = Array.isArray(records) ? '_pushMany' : '_pushOne';
      return this[fnc](type, records);
    };

    Store.prototype.find = function find(type, params) {
      var fnc = this._isManyRequest(type, params) ? '_findMany' : '_findByKey';
      return this[fnc](type, params);
    };

    Store.prototype.fetch = function fetch(type, params) {
      var fnc = this._isManyRequest(type, params) ? '_fetchMany' : '_fetchByKey';
      return this[fnc](type, params);
    };

    Store.prototype.onTransformOne = function onTransformOne(fnc) {
      this._transforms[0].push(fnc);
    };

    Store.prototype.onTransformMany = function onTransformMany(fnc) {
      this._transforms[1].push(fnc);
    };

    Store.prototype.transform = function transform(type, data, isMany) {
      var transforms = this._transforms[~ ~isMany];
      transforms.forEach(function (fnc) {
        data = fnc(type, data);
      });
      return data;
    };

    /**
     * Private
     */

    Store.prototype._isManyRequest = function _isManyRequest(type, params) {
      var typeofParams = typeof params === 'undefined' ? 'undefined' : babelHelpers.typeof(params);
      return typeofParams === 'undefined' || typeofParams === 'object' && !params[this._primaryKeyForType(type)];
    };

    Store.prototype._primaryKeyForType = function _primaryKeyForType(type) {
      var typeOpts = this.lookup(type);
      return typeOpts && typeOpts.primaryKey || this.primaryKey;
    };

    Store.prototype._getByKey = function _getByKey(type, key) {
      var data = this._data[type];
      var record = data && data[key];
      if (record) {
        return clone(record);
      }
    };

    Store.prototype._getMany = function _getMany(type, params) {
      var _this = this;

      var data = this._data[type];
      if (data) {
        var keys = Object.keys(data);
        return keys.length && keys.map(function (key) {
          return _this._getByKey(type, key);
        }) || [];
      }
    };

    Store.prototype._pushOne = function _pushOne(type, record) {
      record = this.transform(type, record);
      if (record) {
        var key = record[this._primaryKeyForType(type)];
        if (typeof key !== 'undefined') {
          key = key.toString();
          this._data[type][key] = record;
          return this._getByKey(type, key);
        }
      }
    };

    Store.prototype._pushMany = function _pushMany(type, records) {
      var _this2 = this;

      records = this.transform(type, records, true);
      if (records) {
        return records.length && records.map(function (record) {
          return _this2._pushOne(type, record);
        }) || [];
      }
    };

    Store.prototype._findByKey = function _findByKey(type, key) {
      var cached = this._getByKey(type, key);
      return cached && Promise.resolve(cached) || this._fetchByKey(type, key);
    };

    Store.prototype._findMany = function _findMany(type, params) {
      var cached = this._getMany(type, params);
      if (cached && cached.length) {
        return Promise.resolve(cached);
      }
      return this._fetchMany(type, params);
    };

    Store.prototype._fetchByKey = function _fetchByKey(type, key) {
      var _this3 = this;

      return this.adapter.get(type, key).then(function (data) {
        return _this3._pushOne(type, data);
      });
    };

    Store.prototype._fetchMany = function _fetchMany(type, params) {
      var _this4 = this;

      return this.adapter.get(type, params).then(function (data) {
        return _this4._pushMany(type, data);
      });
    };

    return Store;
  }();

  exports.Store = Store;
  exports.RESTAdapter = RESTAdapter;
  exports.FetchAdapter = FetchAdapter;
  exports.XHRAdapter = XHRAdapter;
  exports.HTTPAdapter = HTTPAdapter;

}));